using System;
using System.Collections;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Windows.Forms;
using NdisMonitor;
using System.Reflection;
using System.IO;
using System.Text;

namespace NmStandardExtension
{
	/// <summary>
	/// Summary description for NmStandardExtension.
	/// </summary>
	[ Extension( "Standard" ) ]
	public class NmStandardExtension : System.Windows.Forms.UserControl, INmExtension
	{
		private System.ComponentModel.IContainer components;

		public NmStandardExtension()
		{
			LoadScriptText ();

			// This call is required by the Windows.Forms Form Designer.
			InitializeComponent();

			// TODO: Add any initialization after the InitComponent call

		}

		/// <summary>
		/// Clean up any resources being used.
		/// </summary>
		protected override void Dispose( bool disposing )
		{
			if( disposing )
			{
				if( components != null )
					components.Dispose();
			}
			base.Dispose( disposing );
		}

		//
		// # data #
		//

		LogFn							_logFn = null;
		private System.Windows.Forms.ListView ctrlView;
		private System.Windows.Forms.ColumnHeader chOrdinal;
		private System.Windows.Forms.ColumnHeader chTime;
		private System.Windows.Forms.ColumnHeader chSrcDstMAC;
		private System.Windows.Forms.ColumnHeader chSrcDstOther;
		private System.Windows.Forms.ColumnHeader chDesc;
		private System.Windows.Forms.ImageList ilListViewImages;
		NdisHookStubs.NT_PROTOCOL_LIST	_protList = null; // ### WARNING ### this can be NULL !!!

		protected string				_scriptText = null;
		protected Assembly				_scriptAssembly = null;
		protected Type					_scriptClassType = null;
		protected object				_scriptClassObject = null;
		protected MethodInfo			_scriptProcessNextPacketFn = null;
		protected MethodInfo			_scriptUserCommandFn = null;

		//
		// # INmExtension implementation #
		//

		protected class ListViewItemData
		{
			//
			// constructor: same prototype as ProcessNextPacket.
			//

			public ListViewItemData( RawPacket rp, NdisHookStubs.NEXT_PACKET np, int ord, DateTime tm )
			{
				this._rp = rp;
				this._np = np;
				this._ord = ord;
				this._tm = tm;
			}

			//
			// data.
			//

			public RawPacket					_rp;
			public NdisHookStubs.NEXT_PACKET	_np;
			public int							_ord;
			public DateTime						_tm;
		}

		public bool ProcessNextPacket( RawPacket rp, NdisHookStubs.NEXT_PACKET np, int ord, DateTime tm )
		{
			if ( _scriptProcessNextPacketFn == null )
			{
				return ProcessNextPacketImpl( rp, np, ord, tm );
			}
			else
			{
				try
				{
					return (bool) _scriptProcessNextPacketFn.Invoke(
						_scriptClassObject,
						new object[ 6 ] {
											new ProcessNextPacketFn( ProcessNextPacketImpl ), _logFn,
											rp, np, ord, tm } );
				}
				catch( Exception ex )
				{
					// log.
					_logFn( "Exception fired when executing \"ProcessNextPacket\" user implementation. Packet discarded..." );
					_logFn( "=== EXCEPTION DETAILS ===" );
					_logFn( ex.ToString() );

					// return.
					return false;
				}
			}
		}

		protected bool ProcessNextPacketImpl( RawPacket rp, NdisHookStubs.NEXT_PACKET np, int ord, DateTime tm )
		{
			// get the row strings.

			Ethernet_Base			macH = ( rp is EthernetPacket ) ? ((EthernetPacket) rp)._macHeader : null;
			NetProt_IP				ipH = ( rp is EthernetPacket ) && ((EthernetPacket) rp)._netHeader != null && ((EthernetPacket) rp)._netHeader is NetProt_IP ? (NetProt_IP) ((EthernetPacket) rp)._netHeader : null;
			Transport_Protocol		tranH = ( rp is EthernetPacket ) ? ((EthernetPacket) rp)._tranHeader : null;

			string					packetDesc = rp._data.Length.ToString() + " byte(s).";
			if ( ipH != null )
			{
				// print informations about the packet.

				if ( tranH == null || tranH is TranProt_Undefined )
				{
					packetDesc += " Ip protocol = \"" + ipH._protocol.ToString () + "\".";
				}
				else
				{
					if ( tranH is TranProt_ICMP )
						packetDesc += " ICMP ( type = \"" + ((TranProt_ICMP)tranH)._type.ToString () + "\" ).";
					else if ( tranH is TranProt_TCP )
						packetDesc += " TCP ( dstPort = " + ((TranProt_TCP)tranH)._dstPort.ToString() + " <- srcPort = " + ((TranProt_TCP)tranH)._srcPort.ToString() + " ).";
					else if ( tranH is TranProt_UDP )
						packetDesc += " UDP ( dstPort = " + ((TranProt_UDP)tranH)._dstPort.ToString() + " <- srcPort = " + ((TranProt_UDP)tranH)._srcPort.ToString() + " ).";
				}
			}
			else
			{
				// simply print the ETHERTYPE of the protocol.

				UInt16			eType = 0;

				if ( macH is Ethernet_II )
					eType = ( (Ethernet_II) macH )._eType;
				else if ( macH is Ethernet_SNAP )
					eType = ( (Ethernet_SNAP) macH )._ethernetIIType;

				if ( eType != 0 )
					packetDesc += " Ethertype = 0x" + eType.ToString( "X" ) + ".";
			}

			string[]				si = new string[ 5 ];
			si[ 0 ] = "/" + ord.ToString ();
			si[ 1 ] = tm.ToLongTimeString();
			si[ 2 ] = macH != null ? ( np._bDirection==0 ? macH.MacSrcToString() : macH.MacDstToString() ) : "???";
			si[ 3 ] = ipH != null ? ( np._bDirection==0 ? ipH.SrcIPToString() : ipH.DstIPToString() ) : "???";
			si[ 4 ] = packetDesc;

			// add the row.

			ListViewItem			lvi = new ListViewItem(
				si,
				( np._bDirection == 0 ) ? 0 : 1 );
			lvi.Tag = new ListViewItemData( rp, np, ord, tm );

			ctrlView.Items.Add( lvi );

			if ( ctrlView.SelectedItems != null && ctrlView.SelectedItems.Count == 1 )
			{
				Rectangle		rectItem = ctrlView.SelectedItems[ 0 ].Bounds;
				Rectangle		rectCtrl = ctrlView.ClientRectangle;
				if ( rectCtrl.IntersectsWith( rectItem ) == false )
					ctrlView.SelectedItems[ 0 ].Selected = false;
			}

			if ( ctrlView.SelectedItems == null || ctrlView.SelectedItems.Count == 0 )
				ctrlView.EnsureVisible( ctrlView.Items.Count - 1 );

			// return.

			return true;
		}

		protected void CodeWindowClickHandler( object sender, System.EventArgs e )
		{
			// show the window.

			CodeWindow			dlg = new CodeWindow ( _scriptText );
			dlg.ShowDialog ();

			_scriptText = dlg._scriptText;

			if ( dlg._modified )
			{
				// save the reference to the assembly.

				_scriptAssembly = dlg._retval;

				// get the type.

				_scriptClassType = null;
				_scriptClassObject = null;
				_scriptProcessNextPacketFn = null;
				_scriptUserCommandFn = null;

				if ( _scriptAssembly != null )
					_scriptClassType = _scriptAssembly.GetType( "UserHook" );

				if ( _scriptClassType != null )
				{
					try
					{
						_scriptClassObject = Activator.CreateInstance( _scriptClassType );
					}
					catch( Exception ex )
					{
						_logFn( "Exception fired when trying to create an instance of the \"UserHook\" class..." );
						_logFn( "=== EXCEPTION DETAILS ===" );
						_logFn( ex.ToString() );
					}
				}

				if ( _scriptClassObject != null )
				{
					_scriptProcessNextPacketFn = _scriptClassType.GetMethod( "ProcessNextPacket" );
					_scriptUserCommandFn = _scriptClassType.GetMethod( "UserCommand" );
				}
			}

			// save the script.

			SaveScriptText ();

			// return.

			return;
		}

		public void Showing ( LogFn logFn, NdisHookStubs.NT_PROTOCOL_LIST protList, MenuItem popupMenu )
		{
			_logFn = logFn;
			_protList = protList;

			// add menu.

			MenuItem			mi = new MenuItem ();
			mi.Index = 0;
			mi.Text = "&Code window...";
			mi.Click += new System.EventHandler( CodeWindowClickHandler );

			popupMenu.MenuItems.Add( 0, mi );

			// log.

			_logFn( "Standard extension v1.00 for NDIS Monitor .NET." );
		}

		public void Hiding ()
		{
		}

		public void DeleteAllPackets ()
		{
			ctrlView.Items.Clear ();
		}

		public new void Disposing ()
		{
			// show a warning message.

			if ( _scriptProcessNextPacketFn != null )
				MessageBox.Show( "NDIS Monitor is resetting the \"Standard Extension\" state.\n\nYou have successfully compiled a filter logic script in the previous session: the old script will be saved on disk and will be restored automatically in the \"Code Window\" dialog. However YOU NEED TO RECOMPILE IT EXPLICITLY from the \"Code Window\" (pressing Alt+Enter) in order to apply and enforce its filtering logic again in the next capturing session...", "Script WARNING" );

			// FREE EVERYTHING.

			ctrlView.Items.Clear ();
			_protList = null;
		}

		public bool UserCommand( string s )
		{
			if ( _scriptUserCommandFn == null )
			{
				return UserCommandImpl( s );
			}
			else
			{
				try
				{
					return (bool) _scriptUserCommandFn.Invoke(
						_scriptClassObject,
						new object[ 3 ] {
											new UserCommandFn( UserCommandImpl ), _logFn,
											s } );
				}
				catch( Exception ex )
				{
					// log.
					_logFn( "Exception fired when executing \"UserCommand\" user implementation. Command discarded..." );
					_logFn( "=== EXCEPTION DETAILS ===" );
					_logFn( ex.ToString() );

					// return.
					return false;
				}
			}
		}

		protected bool UserCommandImpl( string s )
		{
			return false;
		}

		#region Component Designer generated code
		/// <summary>
		/// Required method for Designer support - do not modify 
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
			this.components = new System.ComponentModel.Container();
			System.Resources.ResourceManager resources = new System.Resources.ResourceManager(typeof(NmStandardExtension));
			this.ctrlView = new System.Windows.Forms.ListView();
			this.chOrdinal = new System.Windows.Forms.ColumnHeader();
			this.chTime = new System.Windows.Forms.ColumnHeader();
			this.chSrcDstMAC = new System.Windows.Forms.ColumnHeader();
			this.chSrcDstOther = new System.Windows.Forms.ColumnHeader();
			this.chDesc = new System.Windows.Forms.ColumnHeader();
			this.ilListViewImages = new System.Windows.Forms.ImageList(this.components);
			this.SuspendLayout();
			// 
			// ctrlView
			// 
			this.ctrlView.Columns.AddRange(new System.Windows.Forms.ColumnHeader[] {
																					   this.chOrdinal,
																					   this.chTime,
																					   this.chSrcDstMAC,
																					   this.chSrcDstOther,
																					   this.chDesc});
			this.ctrlView.Dock = System.Windows.Forms.DockStyle.Fill;
			this.ctrlView.Font = new System.Drawing.Font("Courier New", 8.25F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((System.Byte)(0)));
			this.ctrlView.FullRowSelect = true;
			this.ctrlView.HideSelection = false;
			this.ctrlView.Location = new System.Drawing.Point(0, 0);
			this.ctrlView.MultiSelect = false;
			this.ctrlView.Name = "ctrlView";
			this.ctrlView.Size = new System.Drawing.Size(544, 144);
			this.ctrlView.SmallImageList = this.ilListViewImages;
			this.ctrlView.TabIndex = 0;
			this.ctrlView.View = System.Windows.Forms.View.Details;
			this.ctrlView.KeyDown += new System.Windows.Forms.KeyEventHandler(this.ctrlView_KeyDown);
			this.ctrlView.DoubleClick += new System.EventHandler(this.ctrlView_DoubleClick);
			// 
			// chOrdinal
			// 
			this.chOrdinal.Text = "DIR/#";
			this.chOrdinal.Width = 65;
			// 
			// chTime
			// 
			this.chTime.Text = "Time";
			this.chTime.Width = 72;
			// 
			// chSrcDstMAC
			// 
			this.chSrcDstMAC.Text = "Src/Dst MAC";
			this.chSrcDstMAC.Width = 110;
			// 
			// chSrcDstOther
			// 
			this.chSrcDstOther.Text = "Src/Dst Addr";
			this.chSrcDstOther.Width = 125;
			// 
			// chDesc
			// 
			this.chDesc.Text = "Description";
			this.chDesc.Width = 500;
			// 
			// ilListViewImages
			// 
			this.ilListViewImages.ImageSize = new System.Drawing.Size(16, 16);
			this.ilListViewImages.ImageStream = ((System.Windows.Forms.ImageListStreamer)(resources.GetObject("ilListViewImages.ImageStream")));
			this.ilListViewImages.TransparentColor = System.Drawing.Color.Transparent;
			// 
			// NmStandardExtension
			// 
			this.Controls.Add(this.ctrlView);
			this.Name = "NmStandardExtension";
			this.Size = new System.Drawing.Size(544, 144);
			this.ResumeLayout(false);

		}
		#endregion

		private void ctrlView_DoubleClick(object sender, System.EventArgs e)
		{
			if ( ctrlView.SelectedItems != null && ctrlView.SelectedItems.Count == 1 )
			{
				ListViewItem		sel = ctrlView.SelectedItems[ 0 ];
				ListViewItemData	data = (ListViewItemData) sel.Tag;

				// show hex viewer.

				dlgPacketViewer		dlg = new dlgPacketViewer( data._rp, new dlgPacketViewer.GetPrevOrNextPacketDelegate( GetPrevOrNextPacket ) );
				dlg.ShowDialog();
			}
		}

		protected RawPacket GetPrevOrNextPacket( int dir )
		{
			// special case.
			if ( ctrlView.Items.Count != 0 && ctrlView.SelectedItems != null && ctrlView.SelectedItems.Count == 0 )
				ctrlView.Items[ ctrlView.Items.Count - 1 ].Selected = true;

			// check id.
			if ( ctrlView.SelectedItems != null && ctrlView.SelectedItems.Count == 1 )
			{
				bool		display = false;

				// modify the selection.
				int			id = ctrlView.SelectedItems[ 0 ].Index;
				if ( dir == 0 ) // PREVIOUS
				{
					if ( id != 0 )
					{
						ctrlView.Items[ id - 1 ].Selected = true;
						display = true;
					}
				}
				else // NEXT
				{
					if ( id < ctrlView.Items.Count - 1 )
					{
						ctrlView.Items[ id + 1 ].Selected = true;
						display = true;
					}
				}

				// check whether returning an other packet.
				if ( display )
					return ( (ListViewItemData) ctrlView.SelectedItems[ 0 ].Tag )._rp;
			}

			// return failure.
			return null;
		}

		private void ctrlView_KeyDown(object sender, System.Windows.Forms.KeyEventArgs e)
		{
			if ( e.KeyCode == Keys.Enter )
				ctrlView_DoubleClick( null, null );
		}

		protected string ScriptFileName
		{
			get
			{
				// return.
				string		path = Path.GetDirectoryName( Application.ExecutablePath );
				if ( path[ path.Length - 1 ] != '\\' && path[ path.Length - 1 ] != '/' )
					path += "\\";
				path += "StandardExtScript.cs";
				return path;
			}
		}

		protected void LoadScriptText ()
		{
			_scriptText = "";

			StreamReader	sr = null;
			try
			{
				sr = new StreamReader( ScriptFileName, Encoding.ASCII );
				_scriptText = sr.ReadToEnd ();
			}
			catch
			{
			}
			finally
			{
				if ( sr != null )
					sr.Close ();
			}
		}

		protected void SaveScriptText ()
		{
			StreamWriter		sw = null;
			try
			{
				sw = new StreamWriter( ScriptFileName, false, Encoding.ASCII );
				sw.Write( _scriptText );
			}
			catch
			{
			}
			finally
			{
				if ( sw != null )
					sw.Close ();
			}
		}
	}
}
